/*


    ========== licence begin LGPL
    Copyright (C) 2002 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/

package com.sap.dbtech.jdbcext;

import java.sql.*;
import java.util.*;
import java.io.*;
import javax.sql.*;
import java.lang.ref.*;

import com.sap.dbtech.jdbc.*;

/**
 * This class implements a pooled connection for SAP DB. A pooled connection
 * is represents a <i>a single physical connection</i>. The instance of the 
 * pooled connection is cached by the application server, 
 */
public class PooledConnectionSapDB
    extends ConnectionListenerBag
    implements PooledConnection
{
    Connection physicalConnection;
    WeakReference lastClientConnection;
    
    PooledConnectionSapDB(Connection physicalConnection)
    {
        this.physicalConnection=physicalConnection;
        this.lastClientConnection=null;
    }

    private void sendErrorEvent(SQLException sqlEx)
    {
        super.sendErrorEvent(this, sqlEx);
    }
    
    private void sendCloseEvent()
    {
        super.sendCloseEvent(this);
    }
    
    private PooledConnectionSapDB self()
    { 
        return this; 
    }

    public synchronized Connection getConnection()
        throws SQLException
    {
        // at first look for the weak reference whether
        // we have to close the client connection (forced close).
        Object lastClient=null;
        if(this.lastClientConnection!=null && ((lastClient=this.lastClientConnection.get())!=null)) {
            PooledClientConnectionSapDB pcc=(PooledClientConnectionSapDB)lastClient;
            ConnectionSapDB c=null;
            try {
                c = (ConnectionSapDB)physicalConnection;
            } catch(ClassCastException ccx) {
                // TODO: make this work if trace is enabled.
            }
            if(c!=null) {
                Iterator i=c.getOpenStatements().getall();
                while(i.hasNext()) {
                    Statement s=(Statement)i.next();
                    s.close(); // this behaves well, and throws no exceptions
                }
                c.reinitialize();
            }
            pcc.rollback();
            pcc.closed=true;
            // no close event, though
        }
        
        PooledClientConnectionSapDB result=new PooledClientConnectionSapDB();
        lastClientConnection = new WeakReference(result);
        return result;
    }
    
    public synchronized void close()
        throws SQLException
    {
        physicalConnection.close(); // client connection will suffer automatically
    }

    /**
     * The client connection that is handed out to the application.
     * It will use the physical connection for operation.
     */
    private class PooledClientConnectionSapDB 
        extends ClientConnectionSapDB
    {
        boolean closed=false;
        
        public Connection getPhysicalConnection()
        {
            return physicalConnection;
        }

        public void exceptionOccurred(SQLException sqlEx)
        {
            sendErrorEvent(sqlEx);
        }
        
        private ConnectionSapDB getSapDBConnection(Connection c)
        {
            ConnectionSapDB result=null;
            try {
                result=(ConnectionSapDB)physicalConnection;
                return result;
            } catch(ClassCastException ccx) {
                try {
                    com.sap.dbtech.jdbc.trace.Connection tc=(com.sap.dbtech.jdbc.trace.Connection)physicalConnection;
                    result=(ConnectionSapDB)tc.getInner();
                    return result;
                } catch(ClassCastException ccx2) {
                    return null;
                }
            }
        }

        public void close()
            throws SQLException
        {
            // do not send unnecessary events
            if(closed) {
                return;
            } 
            synchronized(self()) {
                // ... close the open statements ...
                
                ConnectionSapDB c=getSapDBConnection(physicalConnection);
                if(c!=null) {
                    Iterator i=c.getOpenStatements().getall();
                    while(i.hasNext()) {
                        Statement s=(Statement)i.next();
                        s.close(); // this behaves well, and throws no exceptions
                    }
                    
                    c.reinitialize();
                }
                // make a rollback if the autocommit is off.
                if(!getAutoCommit()) {
                    this.rollback();
                }
                this.closed=true;
                sendCloseEvent();
            }
        }
        
        public boolean isClosed()
            throws SQLException
        {
            if(this.closed) {
                return true;
            } else {
                return super.isClosed();
            }
        }

        public void finalize()
        {
            try {
                this.close();
            } catch(SQLException sqlEx) {
                
            }
        }
    }
}
